import java.awt.*;
import java.awt.event.*;
import java.io.*;

import com.threed.jpct.*;

/**
  This application shows some basic and some more advanced usage of jPCT.
  It demonstrates how to create an Object3D without using a Loader,
  how to do simple collision detection and some Camera-stuff.
  Furthermore, it shows how to initialize OpenGL support.
  The movement and animation in this application are NOT framerate independent!
 */
public class TerrainGL {

   /*
     This defines the size of the terrain in "tiles". Do not set this value too high as some
     startup calculations may take ages then.
    */
   final static int X_SIZE=50;
   final static int Z_SIZE=50;

   static boolean software=false;
   static boolean fullscreen=false;

   static int cbpp=16;
   static int zbpp=16;
   static int tbpp=16;

   /**
    * The framecounter
    */
   int frameCounter=0;

   /**
    *  The framebuffer. This is were jPCT will render the final image into
    */
   FrameBuffer buffer=null;

   /**
    * The world. This Object will contain the 3D objects and the lightsources
    */
   World theWorld;

   /**
    * The texture-manager. This is required for loading and managing textures
    */
   TextureManager texMan=null;

   /**
    * The dimensions of the FrameBuffer
    */
   int width=640;
   int height=480;

   /**
    * This is where the ID of the lightsource following the "player" will be stored...later...
    */
   int playerLight=0;

   /**
    * The initial position of the sphere (the "player")
    */
   float sphereX=250f;
   float sphereZ=450f;
   float sphereY=-21f;

   /**
    * The Object3D representing the terrain
    */
   Object3D ground=null;

   /**
    * The Object3D representing the sphere
    */
   Object3D sphere=null;

   Frame frame=null;
   Graphics gframe=null;

   boolean exit=false;

   private TerrainGL() {

      Config.maxPolysVisible=20000;

      try {

         frame=new Frame();
         frame.pack();

         if (!software) {
            /**
             * We are using OpenGL via LWJGL...
             */
            frame.setSize(100, 60);
            frame.setLocation(width, height-60);

            /**
             * We have to find a supported video-mode for OpenGL. This is a simple approach of
             * doing this...there are more intelligent ways of dealing with none-supported
             * video-modes than just exiting...anyway, i'll leave this up to you.
             */
            boolean modeFound=false;

            Config.glTextureDepth=tbpp;

            /**
             * This is required to ensure that the "desiredMode" and the modes[]
             * are initialized with the same z-buffer depth. FrameBuffer.getVideoModes()
             * sets the z-buffer depth of all returned modes to this value, because
             * LWJGL doesn't return this value from the driver:
             */
            Config.glZBufferDepth=zbpp;

            VideoMode desiredMode=new VideoMode(width, height, cbpp, zbpp, 75);
            VideoMode[] modes=FrameBuffer.getVideoModes(IRenderer.RENDERER_OPENGL);

            for (int i=0; i<modes.length; i++) {
               if (modes[i].equalsExactly(desiredMode)) {
                  Config.glSetDesiredVideoMode(modes[i], fullscreen);
                  modeFound=true;
                  break;
               }
            }

            if (!modeFound) {
               /**
                * If no matching mode could be found until here, we are trying it with
                * weaker requirements.
                */
               desiredMode=new VideoMode(width, height, cbpp, zbpp, 0);

               /**
                * Note: We are using a refreshrate of 0 Hz here, because some videodrivers seem
                * to return 0 all the time. So a mode with a refreshrate with 0 or higher
                * will fit here (which should be any mode).
                */
               for (int i=0; i<modes.length; i++) {
                  if (modes[i].equalsAlmost(desiredMode, false)) {
                     Config.glSetDesiredVideoMode(modes[i], fullscreen);
                     modeFound=true;
                     break;
                  }
               }
            }

            if (!modeFound) {
               System.out.println("Desired video-mode not supported!");
               System.exit(-1);
            }

         } else {

            /**
             * We are using software mode...
             */
             frame.setSize(640, 480);
         }
         frame.setResizable(false);
         frame.addWindowListener(new WindowEvents());
         frame.show();

      } catch (Throwable t) {
         System.out.println("uncaught exception: "+t);
         t.printStackTrace();
      }
      gframe=frame.getGraphics();
   }

   public static void main(String[] args) {
      if (args.length>0) {
         /**
          * TerrainGL supports some parameters...here they are...
          */
         for (int i=0; i<args.length; i++) {
            String par=args[i];
            if (par.equalsIgnoreCase("software")) {
               software=true;
            }
            if (par.equalsIgnoreCase("color32")) {
               cbpp=32;
            }
            if (par.equalsIgnoreCase("zbuffer24")) {
               zbpp=24;
            }
            if (par.equalsIgnoreCase("texture32")) {
               cbpp=32;
            }
            if (par.equalsIgnoreCase("fullscreen")) {
               fullscreen=true;
            }
         }
      }
      TerrainGL tgl=new TerrainGL();
      tgl.init();
      tgl.render();
   }

   /**
    * Doing some initializations...
    */
   public void init() {

      /**
       * This array will hold the terrain data to construct the Object3D from.
       * After that, it's not needed anymore. The terrain in this example
       * is X_SIZE * Z_SIZE tiles in size.
       */
      float[][] terrain=new float[X_SIZE][Z_SIZE];


      /**
       * Configure the engine for this application:
       * This is an outdoor-scene with very low overdraw. Different
       * will most likely have different requirements.
       */
      Config.isIndoor=false;
      Config.doPortalHsr=false;
      Config.zTrick=false;
      Config.useBB=true;
      Config.doSorting=false;

      /**
       * Initialize our FrameBuffer
       */
      buffer=new FrameBuffer(width, height, FrameBuffer.SAMPLINGMODE_NORMAL);

      theWorld=new World();
      texMan=TextureManager.getInstance();

      /**
       * We'll use three lights in this scene. Two are static (light and light2) and
       * one is moving along with the player (playerLight)
       */
      int light=theWorld.addLight(new SimpleVector(150, -105, 400), 15f, 2f, 15f);
      int light2=theWorld.addLight(new SimpleVector(400, -90, 150), 15f, 15f, 2f);
      playerLight=theWorld.addLight(new SimpleVector(sphereX, sphereY, sphereZ), 15f, 15f, 15f);

      /**
       * Set the ambient lightsource to negative values, i.e. the scene is quite dark
       */
      theWorld.setAmbientLight(-60, -60, -60);

      /**
       * "Amplify" the lightsources by 20
       */
      Config.lightMul=20;

      /**
       * Now load the textures and add them to the instance of the TextureManager
       */
      char c=File.separatorChar;
      Texture tex=new Texture("data"+c+"grass.jpg");
      texMan.addTexture("base", tex);
      tex=new Texture("data"+c+"grasschrome.jpg");
      texMan.addTexture("chrome", tex);
      tex=new Texture("data"+c+"blittext.jpg");
      texMan.addTexture("blit", tex);

      /**
       * This is where the terrain-array is filled with some data, i.e. the terrain is build.
       * Keep in mind that this approach to build a terrain is quite simple and slow. But it's
       * easy to understand and the results are not that bad, so....
       */
      for (int x=0; x<X_SIZE; x++) {
         for (int z=0; z<Z_SIZE; z++) {
            terrain[x][z]=-20+(float) Math.random()*40f; // Very simple....
         }
      }

      /**
       * We are now smoothing the terrain heights. This looks better because it removes sharp edges
       * from the terrain.
       */
      for (int x=0; x<X_SIZE-1; x++) {
         for (int z=0; z<Z_SIZE-1; z++) {
            terrain[x][z]=(terrain[x][z]+terrain[x+1][z]+terrain[x][z+1]+terrain[x+1][z+1])/4;
         }
      }

      /**
       * The terrain heights are calculated now. Now we have to build an Object3D from them.
       */
      ground=new Object3D(X_SIZE*Z_SIZE*2);

      /**
       * We have 50 tiles in x and z direction and 2 polygons per tile.
       */
      float xSizeF=(float) X_SIZE;
      float zSizeF=(float) Z_SIZE;

      System.out.print("Building terrain...");
      long t=System.currentTimeMillis();
      for (int x=0; x<X_SIZE-1; x++) {
         for (int z=0; z<Z_SIZE-1; z++) {
            /**
             * We are now taking the heights calculated above and build the actual triangles using this data.
             * The terrain's grid is fixed in x and z direction and so are the texture coordinates. The only part
             * that varies is the height, represented by the data in terrain[][].
             * jPCT automatically takes care of vertex sharing and mesh optimizations (this is why building objects
             * this way isn't blazing fast...but it pays out in the end...:-))
             * The Mesh is build with triangle strips in mind here. The format for these strips is equal to that of
             * OpenGL's triangle strips.
             */

            ground.addTriangle(new SimpleVector(x*10, terrain[x][z], z*10), (x/xSizeF), (z/zSizeF),
                               new SimpleVector((x+1)*10, terrain[x+1][z], z*10), ((x+1)/xSizeF), (z/zSizeF),
                               new SimpleVector(x*10, terrain[x][z+1], (z+1)*10), (x/xSizeF), ((z+1)/zSizeF));

            ground.addTriangle(new SimpleVector(x*10, terrain[x][z+1], (z+1)*10), (x/xSizeF), ((z+1)/zSizeF),
                               new SimpleVector((x+1)*10, terrain[x+1][z], z*10), ((x+1)/xSizeF), (z/zSizeF),
                               new SimpleVector((x+1)*10, terrain[x+1][z+1], (z+1)*10), ((x+1)/xSizeF),
                               ((z+1)/zSizeF));
         }
      }
      System.out.println("done in "+(System.currentTimeMillis()-t)+"ms");

      /**
       * compress the Mesh to free unused space
       */
      ground.getMesh().compress();

      /**
       * set the texture of the terrain
       */
      ground.setTexture("base");

      /**
       * this is required after assigning a new texture to an Object3D
       */
      ground.recreateTextureCoords();

      /**
       * the terrain should be lit using specular lighting
       */
      ground.setSpecularLighting(Object3D.SPECULAR_ENABLED);

      /**
       * and it should act as a potential collider for the player
       */
      ground.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);

      /**
       * the ground may use lazy transformations, because it's not transformed anyway...have a look
       * at the documention of this method for more details about lazy transformations.
       */
      ground.enableLazyTransformations();

      /**
       * the terrain is fine now and can be added to the world
       */
      theWorld.addObject(ground);

      /**
       * Now we need the "player" which is represented by a simple sphere in this example.
       * The sphere is environment mapped and uses an additional color to make it a little brighter.
       */
      sphere=Primitives.getSphere(10, 5);
      sphere.setTexture("chrome");
      sphere.setEnvmapped(Object3D.ENVMAP_ENABLED);
      sphere.setAdditionalColor(new Color(60, 60, 60));
      theWorld.addObject(sphere);

      /*
       We have already build the terrain in a strip-friendly way, but the sphere may be not and
       it doesn't hurt to call this method anyway, so...
       */

      theWorld.createTriangleStrips();

      /**
       *  All objects are done...we can now call buildAllObjects() to prepare them for rendering
       */
      theWorld.buildAllObjects();

      /**
       * The origin of the sphere is set here. This has to take place after "building" the object,
       * because calling build() on any object replaces this value with (0,0,0)
       */
      sphere.setOrigin(new SimpleVector(sphereX, sphereY, sphereZ));

      /**
       * Setting up the Camera is all that it's left.
       */
      theWorld.getCamera().rotateX(-0.25f);
      theWorld.getCamera().rotateY((float) Math.PI/2f);
      theWorld.getCamera().setPosition(new SimpleVector(250, -100, 250));

      Thread.currentThread().setPriority(Thread.MAX_PRIORITY);
   }

   /**
    * Within myUpdate(), the FrameBuffer is drawn to screen and the FPS are drawn.
    * Because they are drawn after blitting the framebuffer, this will flicker...
    */
   private void myUpdate(Graphics g) {

      /**
       * Display the framebuffer. This is either the software rendered buffer or the
       * OpenGL rendered buffer. The OpenGL rendered buffer can't be displayed in a component,
       * so "g" is rather senseless in this case, but it doesn't hurt and will simply be
       * ignored in OpenGL mode. This way, no if-else is required.
       */
      buffer.display(g);

      if (!software) {
         g.setColor(Color.black);
         g.fillRect(0, 0, 110, 60);
      }
      g.setColor(Color.white);
      g.drawString("FPS: "+String.valueOf(frameCounter), 10, 35);
   }

   private void render() {

      Thread.currentThread().setPriority(Thread.NORM_PRIORITY);

      /**
       *  Important: NEVER switch the renderer in another Thread. ALWAYS switch it in the
       *  thread that does the rendering!!!
       */
      if (!software) {
         /**
          * If we are using OpenGL, the renderer will be initialized here...
          */
         buffer.enableRenderer(IRenderer.RENDERER_OPENGL, IRenderer.MODE_OPENGL);
      }
      /**
       * in either case, we are not using the (default) legacy software renderer in
       * this application, so we are disabling it here.
       */
      buffer.disableRenderer(IRenderer.RENDERER_SOFTWARE);

      if (software) {
         /**
          * We want software...we'll get software...but in OpenGL-emulation-mode this time.
          */
         buffer.enableRenderer(IRenderer.RENDERER_SOFTWARE, IRenderer.MODE_OPENGL);
      }

      if (!buffer.isInitialized()) {
         /**
          * If something went wrong either with the framebuffer or the renderer, we'll quit.
          */
         System.out.println("Error initializing the framebuffer...aborting!");
         System.exit(1);
      }

      /**
       * Some helper variables for counting frames and polygons
       */
      int cntFrames=0;
      int lastFrames=0;
      long startTime, endTime;

      /**
       * zDif and xDif are the speeds the player has in these directions in world-units.
       */
      float zDif=-1.5f;
      float xDif=-1.5f;

      startTime=System.currentTimeMillis();

      /**
       * This is our "down"-vector. It's used to "drop" the sphere to the ground.
       */
      SimpleVector down=new SimpleVector(0, 1, 0);

      /**
       * And finally: The main loop
       */
      while (!exit) {

         buffer.clear();

         /**
          * The current position of the player is adjusted
          */
         sphereZ+=zDif;
         sphereX+=xDif;

         /**
          * If the player reaches the borders of the terrain, he shouldn't move any further.
          */
         if (sphereZ<15||sphereZ>485) {
            sphereZ-=zDif;
            zDif=0;
         }

         if (sphereX<15||sphereX>485) {
            sphereX-=xDif;
            xDif=0;
         }

         if (zDif!=0||xDif!=0) {

            /**
             * If the player has moved, we are now doing this:
             * Adjust the camera, so that it's always focused on the player
             * using the lookAt()-method.
             * Then the minimal distance between a (virtual) player 21 units high above
             * the terrain and the terrain itself is calculated. Or in other words:
             * The player is "dropped" from a height of 21 units to the terrain.
             * This is a very simple approach to place the player on the terrain. It's
             * not suitable for all kind of applications but it works fine for this example.
             * The value of 21 is derived from the fact that the highest peak in our terrain
             * can only by 20 units above zero. jPCT's y-axis goes down, so -21 places the
             * player 21 units above zero here.
             */
            SimpleVector tmpPos=new SimpleVector(sphereX, -21, sphereZ);
            theWorld.getCamera().lookAt(tmpPos);

            /**
             * the value of 60 is for optimizing the calculation.
             * Have a look in the documentation for more details.
             */
            float trans=ground.calcMinDistance(tmpPos, down, 60);

            if (Object3D.COLLISION_NONE!=trans) {
               /**
                * If the player can be dropped onto the terrain on this spot,
                * we need to calculate its y-position. This is calculated as
                * the distance -21 and -5 (because of the size of the sphere).
                */
               trans-=26;

               /**
                * Because we are using the translate()-method, we have to work
                * with a relative positioning (relative to the position in the last frame).
                */
               float dif=trans-sphereY;
               sphereY+=dif;

               sphere.translate(xDif, dif, zDif);

               /**
                * And finally, the lightsource needs some adjustment to follow the player.
                */
               tmpPos.y+=dif;
               tmpPos.x+=xDif;
               tmpPos.z+=zDif;
               theWorld.setLightPosition(playerLight, tmpPos);

            } else {
               /**
                * If this happens, the value of 60 in calcMinDistance() above needs adjustment.
                * This should not happen in an unmodified example...
                */
               System.out.println("calcMinDistance() has failed!");
            }
         }

         /**
          * If the player has reached a border, its movement has to be adjusted.
          */
         if (zDif==0) {
            zDif=-1.5f+(float) Math.random()*3f;
            if (zDif<0&&zDif>-1) {
               zDif-=1;
            }
            if (zDif>0&&zDif<1) {
               zDif+=1;
            }
         }
         if (xDif==0) {
            xDif=-1.5f+(float) Math.random()*3f;
            if (xDif<0&&xDif>-1) {
               xDif-=1;
            } // Speedup
            if (xDif>0&&xDif<1) {
               xDif+=1;
            }
         }

         /**
          *  Do all the stuff except the scanline conversion
          */
         theWorld.renderScene(buffer);

         /**
          * Do the scanline conversion (into buffer)
          */
         theWorld.draw(buffer);

         /**
          * If needed: Postprocess the framebuffer and copy back- to frontbuffer
          */
         buffer.update();

         /**
          * This will blit our little message-text into the framebuffer.
          */
         buffer.blit(TextureManager.getInstance().getTexture("blit"), 0, 0, 100, 123, 256, 120,
                     FrameBuffer.TRANSPARENT_BLITTING);

         /**
          * This is just performance counter code:
          */
         cntFrames++;
         endTime=System.currentTimeMillis();
         if (endTime-startTime>1000) {
            frameCounter=cntFrames-lastFrames;
            lastFrames=cntFrames;
            startTime=endTime;
         }

         myUpdate(gframe);
         Thread.yield();
      }
      System.exit(0);
   }

   protected final class WindowEvents extends WindowAdapter {

      public void windowClosing(WindowEvent e) {
         System.out.println("Exiting application...");
         System.out.println("BYE!");
         exit=true;
      }
      public void windowIconified(WindowEvent e) {
      }
      public void windowDeiconified(WindowEvent e) {
      }
   }
}
